import pg from 'pg';

function _createPostgresClient(config) {
  return new pg.Client(config);
}

export default class PostgresNotificationListener {
  #connection = null;
  #connectionCheckInterval = null;
  #connectionCheckTimer = null;
  #createPostgresClient = null;
  #isTerminating = false;
  #moduleName = null;
  #postgres = null;
  #reconnectionDelay = null;
  #subscriptions = new Map();

  constructor(config, options = {}) {
    const {
      connectionCheckInterval = 30 * 1000,
      postgres,
      reconnectionDelay = 10 * 1000
    } = config;
    const {
      createPostgresClient = _createPostgresClient,
      moduleName = 'PostgresNotificationListener'
    } = options;

    this.#connectionCheckInterval = connectionCheckInterval;
    this.#createPostgresClient = createPostgresClient;
    this.#moduleName = moduleName;
    this.#postgres = postgres;
    this.#reconnectionDelay = reconnectionDelay;
  }

  initialize() {
    setImmediate(() => {
      this.#connect();
    });
  }

  subscribe(channel, handler) {
    let handlers = this.#subscriptions.get(channel);

    if (!handlers) {
      handlers = new Set();

      this.#subscriptions.set(channel, handlers);

      if (this.#connection) {
        this.#subscribe(channel);
      }
    }

    handlers.add(handler);

    return () => {
      handlers.delete(handler);

      if (handlers.size < 1) {
        this.#subscriptions.delete(channel);
      }
    };
  }

  async terminate() {
    this.#isTerminating = true;

    if (!this.#connection) {
      return;
    }

    await this.#connection.end();

    this.#connection = null;
  }

  #connect() {
    (async () => {
      try {
        const c = this.#createPostgresClient(this.#postgres);

        await c.connect();

        this.#connection = c;

        for (const channel of this.#subscriptions.keys()) {
          await this.#connection.query(`LISTEN ${channel}`);
        }

        this.#connectionCheckTimer = setInterval(() => {
          if (!this.#connection) {
            clearInterval(this.#connectionCheckTimer);

            return;
          }

          (async () => {
            try {
              await this.#connection.query('SELECT pg_backend_pid()');
            } catch (err) {
              this.#reconnect(err);
            }
          })();
        }, this.#connectionCheckInterval);

        this.#connection.on('notification', ({channel, payload}) => {
          const handlers = this.#subscriptions.get(channel);

          if (!handlers || (handlers.size < 1)) {
            console.warn(
              `[${this.#moduleName}::#connect::this.#connection.on('notification')]`,
              `No handlers for channel "${channel}"`
            );

            return;
          }

          for (const handler of handlers) {
            try {
              handler(payload, channel);
            } catch (err) {
              console.warn(
                `[${this.#moduleName}::#connect::this.#connection.on('notification')]`,
                `Error handling message from channel "${channel}":`,
                err
              );
            }
          }
        });

        this.#connection.on('error', err => {
          console.error(
            `[${this.#moduleName}::#connect::this.#connection.on('error')]`,
            'Error:',
            err
          );
        });

        this.#connection.on('end', () => {
          this.#reconnect();
        });

        console.log(`[${this.#moduleName}::#connect] Connection established`);
      } catch (err) {
        this.#reconnect(err);
      }
    })();
  }

  #reconnect(err) {
    clearInterval(this.#connectionCheckTimer);

    if (this.#isTerminating) {
      return;
    }

    if (err) {
      console.error(
        `[${this.#moduleName}::#reconnect]`,
        `Connection failed or lost. Reconnecting in ${this.#reconnectionDelay} msecs...`,
        'Error:',
        err
      );
    }

    if (this.#connection) {
      this.#connection.end();

      this.#connection = null;
    }

    setTimeout(() => {
      if (this.#isTerminating) {
        return;
      }

      this.#connect();
    }, this.#reconnectionDelay);
  }

  #subscribe(channel) {
    (async () => {
      try {
        await this.#connection.query(`LISTEN ${channel}`);
      } catch (err) {
        this.#reconnect(err);
      }
    })();
  }
}

export function createListener(config, options) {
  const listener = new PostgresNotificationListener(config, options);

  listener.initialize();

  return {
    subscribe: (...args) => listener.subscribe(...args),
    terminate: (...args) => listener.terminate(...args)
  };
}
